% convertFEA2SS.m
% Convert Finite Element Analysis (FEA) modal results to State Space Model
% Import Universal File Format (UFF) or UNV file data from I-DEAS
% 2005-07-24, D. Coyne
% Notes:
% 1) Using an extension of the UFF Read m-file, readuff.m
% 2) This version is written for a simple beam example and is not a
% generalization for any finite element model

[UffDataSets,Info,errmsg] = readuff('simple_beam2.unv');

if Info.nErrors ~= 0
    for ii=1:Info.nErrors
        disp(Info.errorMsgs{ii});
    end
end

% echo header information
iHeader=find(Info.dsTypes==151);
UffDataSets{iHeader}

% echo units information
iUnits=find(Info.dsTypes==164);
UffDataSets{iUnits}

% modal analysis information
iModalAnalysis=find(Info.dsTypes==2414);
UffDataSets{iModalAnalysis}

% form the diagonal matrix of natural frequencies, omega
nModes = length(iModalAnalysis);
freq = zeros(nModes,1);
for ii=1:nModes
    freq(ii) = UffDataSets{iModalAnalysis(ii)}.frequency*2*pi;
end
omega = diag(freq);

% check that every mode has the same nodes
% SHOULD CHECK THAT THE ORDER OF THE NODAL DATA IS THE SAME AS WELL -- PENDING
nodeNumbers = UffDataSets{iModalAnalysis(1)}.nodeNum;
nNodes = length(nodeNumbers);
for ii=2:nModes
    if length(UffDataSets{iModalAnalysis(ii)}.nodeNum) ~= nNodes
        errMessage = ['number of nodes for mode ' num2str(iModalAnalysis(ii).modeNumber) ' does not match mode ' num2str(iModalAnalysis(1).modeNumber)];
        return
    end
    if length(setdiff(UffDataSets{iModalAnalysis(ii)}.nodeNum,nodeNumbers)) ~= 0
        errMessage = ['node numbers for mode ' num2str(iModalAnalysis(ii).modeNumber) ' does not match mode ' num2str(iModalAnalysis(1).modeNumber)];
        return
    end
end

% form the modal coefficient matrix (matrix of eigenvectors), phi
% degree-of-freedom (dof) order is {x1, y1, z1, rx1, ry1, rz1, x2, y2, z2,
% rx2, ry2, rz2, x3, ...}
% Note: Assuming (hard coded) that ndval = 6 (x, y, z, rx, ry, rz) and not 3 (x, y, z)
nDof = nNodes*6;
phi = zeros(nDof,nModes);
for ii=1:nModes
    x = UffDataSets{iModalAnalysis(ii)}.r1;
    y = UffDataSets{iModalAnalysis(ii)}.r2;
    z = UffDataSets{iModalAnalysis(ii)}.r3;
    rx = UffDataSets{iModalAnalysis(ii)}.r4;
    ry = UffDataSets{iModalAnalysis(ii)}.r5;
    rz = UffDataSets{iModalAnalysis(ii)}.r6;
    dofs = [x y z rx ry rz]';
    dofs = reshape(dofs,prod(size(dofs)),1);
    phi(:,ii) = dofs;
end

% form the modal mass matrix, Mm
modalMasses = zeros(nModes,1);
for ii=1:nModes
    modalMasses(ii) = UffDataSets{iModalAnalysis(ii)}.modalMass;
end
Mm = diag(modalMasses);

% form the modal input matrix, Bm
% for this example:
% u = the force is at the beam tip, node 2, in the y-direction
B = zeros(nDof,1);
B(1*6+2,1) = 1;
Bm = inv(Mm) * transpose(phi) * B;

% form the modal output matrix, Cm
% for this example:
% y = y-displacement outputs at nodes 2, 5 and 8
% the output velocity matrix, Cv = [0]
% the output displacement matrix, Cq = [0,1,0,0,1,0,0,1,0,0,0]
nOutputs = 3;
Cq = zeros(nOutputs,nDof);
Cq(1,1*6+2) = 1;
Cq(2,4*6+2) = 1;
Cq(3,7*6+2) = 1;
Cmq = Cq * phi;

% modal proportional damping matrix, Z
modalQ = 100;
modalHystereticDampingRatio = 1/(2*modalQ);
Z = diag(modalHystereticDampingRatio*ones(nModes,1));

% Structural Transfer Function Plot
% matches the transfer function plot produced by I-DEAS response analysis
nFreqs = 500;
f = logspace(1,3,nFreqs);
w = 2*pi*f;
G = zeros(3,nFreqs);
% Note: The following structural Transfer Function is eqn. 2.17 from W. Gawronski's
% Dynamics and Control of Structures. However, I found it necessary to change sign
% on the term (2 i w Z omega) in order to match the direction of phase changes in the
% transfer functions from I-DEAS
for ii=1:nFreqs
    G(:,ii) = Cmq * inv(omega^2 - w(ii)^2 * eye(nModes) - 2*sqrt(-1)*w(ii)*Z*omega) * Bm;
end

% define a frequency response data model
Gsysmodal = frd(G',f,'Units','Hz');

% Import I-DEAS response data for comparison with the response derived from
% the modal imported data
[sys] = import_ideasResponseFunc_dataset58('tipForce_dataset58.unv');

figure;
bode(sys(1,1,1),'r',sys(1,1,2),'g',sys(1,1,3),'b',Gsysmodal(1,1,1),'y--',Gsysmodal(1,1,2),'w--',Gsysmodal(1,1,3),'m--');
sysaxes=findall(gcf,'XScale','log');
set(sysaxes,'XScale','linear');
title(['Simple Beam, Tip Actuation']);
legend('node 2 I-DEAS','node 5 I-DEAS','node 8 I-DEAS','node 2 modal','node 5 modal','node 8 modal');
hfig = findall(gcf,'type','line','visible','on');
% lets make the linewidth of the I-DEAS calculated transfer functions wider
% so that the overlap of the lines is visible and set the color to be
% similar but lighter in shade
set(hfig,'linewidth',2.0);
a = findobj(gcf,'Type','axes');
h = findobj(a(2),'Type','line');
set(h(4:6),'linewidth',5.0);
set(h(1),'Color',[.8 .8 1]);
set(h(2),'Color',[.8 1 .8]);
set(h(3),'Color',[1 .8 .8]);
h = findobj(a(3),'Type','line');
set(h(4:6),'linewidth',5.0);
set(h(1),'Color',[.8 .8 1]);
set(h(2),'Color',[.8 1 .8]);
set(h(3),'Color',[1 .8 .8]);


% Transform into a State Space model
% xdot = A x + B u
% y = C x + D u
A = [zeros(nModes) diag(ones(nModes,1));
    -omega^2 -2*Z*omega];
B = [zeros(nModes,1);
    Bm];
C = [Cmq zeros(3,nModes)];
D = zeros(3,1);
sysSS = ss(A,B,C,D);
size(sysSS)
sysSS

figure;
bode(sysSS);
bode(sysSS(1,1),'r',sysSS(2,1),'g',sysSS(3,1),'b');

figure;
bode(sys(1,1,1),'r',sys(1,1,2),'g',sys(1,1,3),'b',sysSS(1,1),'y--',sysSS(2,1),'w--',sysSS(3,1),'m--',w);
sysaxes=findall(gcf,'XScale','log');
set(sysaxes,'XScale','linear','XLim',[10 1000]);
title(['Simple Beam, Tip Actuation']);
legend('node 2 I-DEAS','node 5 I-DEAS','node 8 I-DEAS','node 2 SS','node 5 SS','node 8 SS');
hfig = findall(gcf,'type','line','visible','on');
% lets make the linewidth of the I-DEAS calculated transfer functions wider
% so that the overlap of the lines is visible and set the color to be
% similar but lighter in shade
set(hfig,'linewidth',2.0);
a = findobj(gcf,'Type','axes');
h = findobj(a(2),'Type','line');
set(h(4:6),'linewidth',5.0);
set(h(1),'Color',[.8 .8 1]);
set(h(2),'Color',[.8 1 .8]);
set(h(3),'Color',[1 .8 .8]);
h = findobj(a(3),'Type','line');
set(h(4:6),'linewidth',5.0);
set(h(1),'Color',[.8 .8 1]);
set(h(2),'Color',[.8 1 .8]);
set(h(3),'Color',[1 .8 .8]);

