function [UffDataSets,Info,errmsg] = readuff(fileName)
%READUFF Reads UFF (Universal File Format) files of types:
%   151, 15, 55, 58, 82, 164, and also the hybrid one, 58b
%   [UffDataSets,Info] = readuff(filename)
%
%   - fileName:     name of the uff file to read data from (all the data)
%
%   - UffDataSets:  an array of structures; each structure holds one data set
%                   (the data set between -1 and -1; Each structure,
%                   UffDataSets{i}, has the fields
%                       .dsType
%                       .binary
%                   and some additional field which are data-set dependant and are
%                   as follows:
%                   #58 - for measurement data - function at dof (58)
%                       .d1 (description 1)     .d2 (description 2)     .date
%                       .ID_4                   .ID_5                   .functionType (see notes)
%                       .loadCaseId             .measData               .refEntName
%                       .refDir                 .refNode                .rspDir
%                       .rspEntName             .rspNode                .x (time or frequency)
%                       .dx (abscisa spacing)   
%
%                   #58b - for measurement data - the same as 58 but the data
%                   is written in binary format
%
%                   #15 - coordinate data (15)  (Grid points):
%                       .nodeN                  .defCS                  .dispCS
%                       .color                  .x                      .y
%                       .z
%
%                   #82 - display Sequence data (82):
%                       .traceNum               .nNodes                 .color
%                       .ID                     .lines
%
%                   #151 - header data (151):
%                       .modelName              .description            .dateCreated
%                       .timeCreated            .dateSaved              .timeSaved
%                       .dbApp                  .dbVersion              .uffApp
%
%                   #164 - units (164):
%                       .unitsCode              .unitsDescription       .tempMode (1=absoulute, 2=relative)
%                       Unit factors for converting universal file units to SI. To convert from
%                       universal file units to SI divide by the appropriate factor listed below:
%                       .facLength              .facForce               .facTemp            
%                       .facTempOffset
%
%                   #55 - data at nodes (55):
%                       /Common fields:/
%                       .analysisType           .dataCharacter = 1      .r1
%                       .r2                     .r3                     .responseType
%                       .r4                     .r5                     .r6
%                       /Normal modes specific fields (analysisType = 2)/
%                       .modeNum                .modeFreq               .modeMass 
%                       .mode_v_damping_ratio   .mode_h_damping_ratio                   
%                       /...or, for complex modes specific fields (analysisType = 3 or 7)/
%                       .modeNum                .eigVal                 .modalA        
%                       .modalB                 
%                       /...or, for frequency response specific fields (analysisType = 5)/
%                       .freqNum                .freq
%
%   - Info:         (optional) structure with the following fields:
%                   .dsTypes    -   an array of data-set types read
%                   .binary     -   an array of 1s (binary format) and 0s (ascii format)
%                   .nDataSets  -   number of data sets found
%                   .errcode    -   an array of error codes for each data
%                                   set; 0 = no error otherwise an error occured in data
%                                   set read - see errmsg
%                   .errmsg     -   error messages (cell array of strings) for each
%                                   data set - empty if no error occured at specific data set
%                   .nErrors    -   number of errors found (unsupported
%                                   datasets, error reading data set,...)
%                   .errorMsgs  -   all the error messages (empty if no error is found)
%   - errmsg:       (optional) general (overall), file-based error
%                   messages - to enable reading of uncorrupted data for
%                   example,...
%
%
%   NOTES: r1..r6 are response vectors with node numbers in ROWS and
%   direction in COLUMN (r1=x, r2=y,...,r6=rz).
%
%   functionType can be one of the following:
%               0 -  General or Unknown
%               1 -  Time Response
%               2 -  (supported) Auto Spectrum
%               3 -  (supported) Cross Spectrum
%               4 -  (supported) Frequency Response Function
%               5 -  Transmissibility
%               6 -  (supported) Coherence
%               7 -  Auto Correlation
%               8 -  Cross Correlation
%               9 -  Power Spectral Density (PSD)
%               10 - Energy Spectral Density (ESD)
%               11 - Probability Density Function
%               12 - Spectrum
%               13 - Cumulative Frequency Distribution
%               14 - Peaks Valley
%               15 - Stress/Cycles
%               16 - Strain/Cycles
%               17 - Orbit
%               18 - Mode Indicator Function
%               19 - Force Pattern
%               20 - Partial Power
%               21 - Partial Coherence
%               22 - Eigenvalue
%               23 - Eigenvector
%               24 - Shock Response Spectrum
%               25 - Finite Impulse Response Filter
%               26 - Multiple Coherence
%               27 - Order Function
%
%   analysisType can be one of the following:
%               0: Unknown
%               1: Static
%               2: (supported) Normal Mode
%               3: (supported) Complex eigenvalue first order
%               4: Transient
%               5: (supported) Frequency Response
%               6: Buckling
%               7: (supported) Complex eigenvalue second order
%
%   dataCharacter can be one of the following:
%               0: Unknown
%               1: Scalar
%               2: 3 DOF Global Translation Vector
%               3: 6 DOF Global Translation & Rotation Vector
%               4: Symmetric Global Tensor
%
%   unitsCode can be one of the following:
%               1 - SI: Meter (newton)
%               2 - BG: Foot (pound f)
%               3 - MG: Meter (kilogram f)
%               4 - BA: Foot (poundal)
%               5 - MM: mm (milli newton)
%               6 - CM: cm (centi newton)
%               7 - IN: Inch (pound f)
%               8 - GM: mm (kilogram f)
%
%   functionType as well as other parameters are described in
%   Test_Universal_File_Formats.pdf
%
%   See also: WRITEUFF
%
%   SOURCES:    [1] Bryce Gardner's read_uff obtained from the internet
%               [2] http://www.sdrl.uc.edu/uff/SDRChelp/LANG/English/unv_ug/book.htm
%
%
%   FREE SOFTWARE - please refer the source
%   Copyright (c) 2004-2005 by Primoz Cermelj
%   First release on 30.05.2004
%   Primoz Cermelj, Slovenia
%   Contact: primoz.cermelj@email.si
%   Download location: http://www.mathworks.com/matlabcentral/fileexchange/loadFile.do?objectId=6395&objectType=file
%
%   Version: 0.9.8b1
%   Last revision: 06.06.2005
%
%   Bug reports, questions, etc. can be sent to the e-mail given above.
%
%   This programme is free software; you can redistribute it and/or
%   modify it under the terms of the GNU General Public License
%   as published by the Free Software Foundation; either version 2
%   of the License, or any later version.
%
%   This programme is distributed in the hope that it will be useful,
%   but WITHOUT ANY WARRANTY; without even the implied warranty of
%   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%   GNU General Public License for more details.
%--------------------------------------------------------------------------

%----------------
% READUFF history
%----------------
%
% [v.0.9.8b1] 06.06.2005
% - NEW: hybrid binary-58 format (58b) is now supported
% - NEW: binary field was added to UffDataSets structures
% [v.0.9.7] 24.05.2005
% - NEW: dsType field was added to UffDataSets structures
% [v.0.9.6b4] 11.05.2005
% - FIX: Matlab version down to 5.3 is now supported
% - FIX: Some minor bugs removed
% - NEW: Speed improvement; reading is much faster now
% 2005-07-15 D. Coyne Correction for end of block marker
% 2005-07-15 D. Coyne Added transposes to blocks to match dimensions
% 2005-07-15 D. Coyne Added function extract2414
%
%----------------

%--------------
% Default outputs
%--------------
UffDataSets = [];
Info.errcode = [];
Info.nDataSets = 0;
Info.dsTypes = [];
Info.binary = [];
Info.errmsg = [];
Info.nErrors = 0;
errmsg = [];

%--------------
% Some variables
%--------------
errN = 0;               % current global error number (data-set number independant)

%--------------
% Read whole file data into an array of characters
%--------------
try
    fid = fopen(fileName,'r');
    if fid == -1,
        errN = errN + 1;
        errmsg{errN,1} = ['could not open file: ' fileName];
        disp(errmsg{errN});
        return
    end
    FILE_DATA = (fread(fid,'char=>char')).';
catch
    errN = errN + 1;
    errmsg{errN,1} = ['error reading file contents: ' lasterr];
    disp(errmsg{errN});
    % Close the file
    fclose(fid);
    return
end
% Close the file
err = fclose(fid);
if err == -1
    errN = errN + 1;
    errmsg{errN,1} = 'error while closing file';
    disp(errmsg{errN});
end

%--------------
% Find all valid blocks, data between -1 and -1; pointers to blocks of
% data; include the first  -1 but exclude the last -1;
% the fist  -1 will be skipped further later on in get_block_prop
%--------------
% 2005-07-15 D. Coyne Correction for end of block marker
%ind = strfind(FILE_DATA,'    -1');
ind = strfind(FILE_DATA,sprintf('    -1\n'));

nBlocks = floor(length(ind)/2);
if nBlocks < 1
    errN = errN + 1;
    errmsg{errN,1} = 'No valid blocks found';
    disp(errmsg{errN});
    return
end
blocks = zeros(nBlocks,2);
% 2005-07-15 D. Coyne Added transposes to match dimensions
blocks(:,1) = ind(1:2:end)';
blocks(:,2) = ind(2:2:end)'-1;

%=============================
% MAIN FILE LOOP - go through all the blocks and extract data from each
% block according to the data type
%=============================
dataSetN = 0;       % counts VALID data-sets (including non-supported ones)
try
    for ii=1:nBlocks
        
        % Skips the first  -1, detects the data-set type and any possible
        % properties (e.g., for 58b there are some additional fields in the data-set
        % id record), and also returnes blockLines - pointers to start and
        % end offsets of lines of the data-set-block data
        [data_set_type,DataSetProp,blockLines,errMessage] = get_block_prop(blocks(ii,1),blocks(ii,2),FILE_DATA);
        if ~isempty(errMessage)
            errN = errN + 1;
            errmsg{errN,1} = errMessage;
            continue
        end
        
        dataSetN = dataSetN + 1;
        if data_set_type == 58      % Function at nodal dof
            [ds_data,ds_errmsg] = extract58(fileName,FILE_DATA,blockLines,DataSetProp);
        elseif data_set_type == 15  % Coordinate data
            [ds_data,ds_errmsg] = extract15(FILE_DATA,blockLines);
        elseif data_set_type == 151 % Header data
            [ds_data,ds_errmsg] = extract151(FILE_DATA,blockLines);
        elseif data_set_type == 164 % Units data
            [ds_data,ds_errmsg] = extract164(FILE_DATA,blockLines);
        elseif data_set_type == 82  % Display Sequence data
            [ds_data,ds_errmsg] = extract82(FILE_DATA,blockLines);
        elseif data_set_type == 55  % Modal data file
            [ds_data,ds_errmsg] = extract55(FILE_DATA,blockLines);
        elseif data_set_type == 2414  % Analysis Mode Shape
            [ds_data,ds_errmsg] = extract2414(FILE_DATA,blockLines);
        else
            ds_data = [];
            ds_errmsg = ['unknown data-set (' num2str(data_set_type)  ') found in ' num2str(ii) '-th data-set '];
        end
        
        UffDataSets{dataSetN} = ds_data;
        UffDataSets{dataSetN}.dsType = data_set_type;
        UffDataSets{dataSetN}.binary = DataSetProp.binary;
        Info.errmsg{dataSetN} = ds_errmsg;
        Info.dsTypes(dataSetN) = data_set_type;
        Info.binary(dataSetN) = DataSetProp.binary;
        if isempty(ds_errmsg)
            Info.errcode(dataSetN) = 0;
        else
            Info.errcode(dataSetN) = 1;
        end
    end
    
catch
    errN = errN + 1;
    errmsg{errN,1} = lasterr;      
end
%=============================
% END OF MAIN FILE LOOP
%=============================

Info.nErrors = length(find(Info.errcode));
Info.nDataSets = dataSetN;
Info.errorMsgs = Info.errmsg(find(Info.errcode));

if ~isempty(errmsg)
    for ii=1:length(errmsg)
        disp(errmsg{ii});
    end
end




%==========================================================================
%                       SUBFUNCTIONS SECTION
%==========================================================================



%--------------------------------------------------------------------------
function [dataSet,DataSetProp,blockLines,errMessage] = get_block_prop(so,eo,FILE_DATA)
% Extract block-data lines' pointers (start and end for each line) and also
% returns the data-set number identified along with any additional
% parameters such as in the case of 58b data-set. so points to the first -1 tag
% (designated by o): o___-1 while eo points to the end -1 tag: o___-1.
% blockLInes are start and end offsets of each line in the current data-set
% starting from the line right after the data-set id line.

% Scans for block data and returns lines' pointers (start and end offsets
% in a 2-column matrix).

errMessage = [];
try
    blockData = FILE_DATA(so:eo);
    dataLen = length(blockData);
    lineBreaks = find(blockData==10 | blockData==13);

    fromInd = ones(length(lineBreaks),1);
    toInd = ones(length(lineBreaks),1);
    for ii=2:length(toInd)
        if (lineBreaks(ii)-1) == lineBreaks(ii-1)
            toInd(ii) = 0;
        end
        if (lineBreaks(ii-1)+1) == lineBreaks(ii)
            fromInd(ii-1) = 0;
        end
    end
    from = ([1 lineBreaks(find(fromInd))+1])';
    to = (lineBreaks(find(toInd))-1)';
    if from(end) >= dataLen
        from(end) = [];
    end
    if (dataLen > to(end)+1) & (lineBreaks(end)~=dataLen)
        to = [to ; dataLen];
    end
    
    % Local blockLines
    blockLines = [from to];

    dataSetLine = blockData(blockLines(2,1):blockLines(2,2));
    dataSet = sscanf(dataSetLine(1:6),'%i',1);
    if isempty(dataSet)
        errMessage = ['invalid data-set type specification found'];
        return
    end
    
    if length(dataSetLine) < 7
        format = '';
    else
        format = sscanf(dataSetLine(7),'%c',1);
    end
    
    if strcmpi(format,'b')
        DataSetProp.binary = 1;
        DataSetProp.byteOrdering = sscanf(dataSetLine(8:13),'%i',1);
        DataSetProp.fpFormat = sscanf(dataSetLine(14:19),'%i',1);
        DataSetProp.nAsciiLines = sscanf(dataSetLine(20:31),'%i',1);
        DataSetProp.nBytes = sscanf(dataSetLine(32:43),'%i',1);
        DataSetProp.d1 = sscanf(dataSetLine(44:49),'%i',1);
        DataSetProp.d2 = sscanf(dataSetLine(50:55),'%i',1);
        DataSetProp.d3 = sscanf(dataSetLine(56:67),'%i',1);
        DataSetProp.d4 = sscanf(dataSetLine(68:end),'%i',1);
    else
        DataSetProp.binary = 0;
    end
    
    % Global blockLines (with respect to FILE_DATA)
    blockLines = blockLines(3:end,:) + so - 1;
    if size(blockLines,1) < 2
        errMessage = ['empty data block found'];
        return
    end
    
catch
    errMessage = ['error reading data-set properties: ' lasterr];
    return
end




%--------------------------------------------------------------------------
function [UFF,errMessage] = extract58(fileName,DATA,blockLines,DataSetProp)
% #58 - Extract data-set type 58 data

UFF = [];
UFF.measData = [];
errMessage = [];
lineN = 1;
nLines = size(blockLines,1);

try
    % Line 1
    UFF.d1 = sscanf(DATA(blockLines(1,1):blockLines(1,2)),'%c',80);
    lineN = lineN + 1;
    % Line 2
    UFF.d2 = sscanf(DATA(blockLines(2,1):blockLines(2,2)),'%c',80);
    lineN = lineN + 1;
    % Line 3
    UFF.date = sscanf(DATA(blockLines(3,1):blockLines(3,2)),'%c',80);
    lineN = lineN + 1;
    % Line 4
    UFF.ID_4 = sscanf(DATA(blockLines(4,1):blockLines(4,2)),'%c',80);
    lineN = lineN + 1;
    % Line 5
    UFF.ID_5 = sscanf(DATA(blockLines(5,1):blockLines(5,2)),'%c',80);
    lineN = lineN + 1;
    % Line 6
    tmpLine = DATA(blockLines(6,1):blockLines(6,2));
    tmpLine = [tmpLine repmat(' ',1,80-length(tmpLine))];
    UFF.functionType = sscanf(tmpLine(1:5),'%i',1);    
    tmp = sscanf(tmpLine(6:15),'%i',1);
    tmp = sscanf(tmpLine(16:20),'%i',1);
    UFF.loadCaseId = sscanf(tmpLine(21:30),'%i',1);
    if UFF.loadCaseId ~= 0,
        errMessage = ['need single point excitation at: ' num2str(lineN)];
        return
    end
    UFF.rspEntName = sscanf(tmpLine(32:41),'%s',1);
    UFF.rspNode = sscanf(tmpLine(42:51),'%i',1);
    UFF.rspDir = sscanf(tmpLine(52:55),'%i',1);
    UFF.refEntName = sscanf(tmpLine(57:66),'%s',1);
    UFF.refNode = sscanf(tmpLine(67:76),'%i',1);
    UFF.refDir = sscanf(tmpLine(77:80),'%i',1);
    lineN = lineN + 1;

    % Line 7; data form
    tmpLine = DATA(blockLines(7,1):blockLines(7,2));
    tmpLine = [tmpLine repmat(' ',1,80-length(tmpLine))];
    ordDataType = sscanf(tmpLine(1:10),'%i',1);
    numpt = sscanf(tmpLine(11:20),'%i',1);
    spacingType = sscanf(tmpLine(21:30),'%i',1);
% 2005-7-21 D. Coyne Extending extract58 to handle uneven abscissa spacing
%    if spacingType ~= 1,
%        errMessage = ['need evenly spaced frequency/time at:' num2str(lineN)];
%        return
%    end
    UFF.xmin = sscanf(tmpLine(31:43),'%g',1);
    UFF.dx = sscanf(tmpLine(44:56),'%g',1);
    zValue = sscanf(tmpLine(57:69),'%g',1);
    lineN = lineN + 1;

    % Line 8; abscissa data characteristics
    tmpLine = DATA(blockLines(8,1):blockLines(8,2));
    tmpLine = [tmpLine repmat(' ',1,80-length(tmpLine))];
    abDataType = sscanf(tmpLine(1:10),'%i',1);
    if ~(abDataType == 18 | abDataType == 17 | abDataType),  % Frequency, time, or unknown
        errMessage = ['need frequency, time, or unknown data on abcissa at:' num2str(lineN)];
        return
    end
    tmp = sscanf(tmpLine(11:15),'%i',1);
    tmp = sscanf(tmpLine(16:20),'%i',1);
    tmp = sscanf(tmpLine(21:25),'%i',1);
    temp = sscanf(tmpLine(27:46),'%s',1);
    temp = sscanf(tmpLine(48:end),'%s',1);
    lineN = lineN + 1;

    % Line 9; Ordinate (or ordinate numerator) Data Characteristics
    tmpLine = DATA(blockLines(9,1):blockLines(9,2));
    tmpLine = [tmpLine repmat(' ',1,80-length(tmpLine))];
    ordNumeratorDataType = sscanf(tmpLine(1:10),'%i',1);
    %  if or_num_data_type(loop) ~= 12,  %Acceleration
    %    ermessage = 'expected acceleration data in the ordinate numerator';
    %    disp(ermessage);
    %  end
    tmp = sscanf(tmpLine(11:15),'%i',1);
    tmp = sscanf(tmpLine(16:20),'%i',1);
    tmp = sscanf(tmpLine(21:25),'%i',1);
    temp = sscanf(tmpLine(27:46),'%s',1);
    temp = sscanf(tmpLine(48:end),'%s',1);
    lineN = lineN + 1;

    % Line 10; Ordinate Denominator Data Characteristics
    tmpLine = DATA(blockLines(10,1):blockLines(10,2));
    tmpLine = [tmpLine repmat(' ',1,80-length(tmpLine))];
    ordDenominatorDataType = sscanf(tmpLine(1:10),'%i',1);
    %  if or_den_data_type(loop) ~= 13,  % Excitation force
    %    ermessage = 'expected force data in the ordinate denominator';
    %    disp(ermessage);
    %  end
    tmp = sscanf(tmpLine(11:15),'%i',1);
    tmp = sscanf(tmpLine(16:20),'%i',1);
    tmp = sscanf(tmpLine(21:25),'%i',1);
    temp = sscanf(tmpLine(27:46),'%s',1);
    temp = sscanf(tmpLine(48:end),'%s',1);
    lineN = lineN + 1;

    % Line 11; Z-axis Data Characteristics
    tmpLine = DATA(blockLines(11,1):blockLines(11,2));
    tmpLine = [tmpLine repmat(' ',1,80-length(tmpLine))];
    tmp = sscanf(tmpLine(1:10),'%i',1);
    tmp = sscanf(tmpLine(11:15),'%i',1);
    tmp = sscanf(tmpLine(16:20),'%i',1);
    tmp = sscanf(tmpLine(21:25),'%i',1);
    temp = sscanf(tmpLine(27:46),'%s',1);
    temp = sscanf(tmpLine(48:end),'%s',1);
    lineN = lineN + 1;

    % Line 12 ...; Data Values
    if DataSetProp.binary
        if DataSetProp.byteOrdering == 1; format = 'l'; else; format = 'b'; end;
        if (ordDataType==2 | ordDataType==5);
            prec = 'float32';
            numLen = 4;
        else
            prec = 'double';
            numLen = 8;
        end
        fid = fopen(fileName,'r',format);
        if fid == -1
            errMessage = ['could not reopen file for binary data reading: ' fileName];
            return
        end
        status = fseek(fid, blockLines(12,1)-1, 'bof');
        if status
            errMessage = ['could not start reading binary data from ' fileName];
            return
        end
        values = fread(fid, round(DataSetProp.nBytes/numLen), prec);
        fclose(fid);
    else
        values = sscanf(DATA(blockLines(12,1):blockLines(end,2)),'%g');
    end

% 2005-7-21 D. Coyne extending to include uneven abscissa
    if spacingType == 1
% even abscissa
        if (ordDataType == 2 | ordDataType == 4)        % real, single or double precision
            UFF.measData = values;
        elseif (ordDataType == 5 | ordDataType == 6)    % complex, single or double precision
            UFF.measData = values(1:2:end-1) + j*values(2:2:end);
        else
            errMessage = ['error reading measurement data at:' num2str(lineN)];
            return
        end
        nVal = length(UFF.measData);
        % Abscissa
        UFF.x = [ UFF.xmin : UFF.dx : (nVal-1)*UFF.dx ];
    else
% uneven abscissa
        if ordDataType == 5    % complex, single precision
            freq = values(1:3:end-2);
            UFF.measData = values(2:3:end-1) + j*values(3:3:end);
        else
            errMessage = ['error reading measurement data at:' num2str(lineN)];
            return
        end
        nVal = length(UFF.measData);
        % Abscissa
        UFF.x = freq;
    end
        
catch
    errMessage = ['error reading measurement data: ' lasterr];
    return
end


%--------------------------------------------------------------------------
function [UFF,errMessage] = extract151(DATA,blockLines)
% #151 - Extract data-set type 151 data

UFF = [];
errMessage = [];
lineN = 1;
nLines = size(blockLines,1);

try
    % Line 1
    UFF.modelName = sscanf(DATA(blockLines(1,1):blockLines(1,2)),'%c',80);
    lineN = lineN + 1;
    % Line 2
    UFF.description = sscanf(DATA(blockLines(2,1):blockLines(2,2)),'%c',80);
    lineN = lineN + 1;
    % Line 3
    UFF.dbApp = sscanf(DATA(blockLines(3,1):blockLines(3,2)),'%c',80);
    lineN = lineN + 1;
    % Line 4
    tmpLine = DATA(blockLines(4,1):blockLines(4,2));
    tmpLine = [tmpLine repmat(' ',1,80-length(tmpLine))];
    UFF.dateCreated = sscanf(tmpLine(1:10),'%s',10);
    UFF.timeCreated = sscanf(tmpLine(11:20),'%s',10);
    UFF.dbVersion = sscanf(tmpLine(21:30),'%i',10);
    lineN = lineN + 1;       
    % Line 5
    tmpLine = DATA(blockLines(5,1):blockLines(5,2));
    tmpLine = [tmpLine repmat(' ',1,80-length(tmpLine))];
    UFF.dateSaved = sscanf(tmpLine(1:10),'%s',10);
    UFF.timeSaved = sscanf(tmpLine(11:20),'%s',10);
    lineN = lineN + 1;       
    % Line 6
    UFF.uffApp = sscanf(DATA(blockLines(6,1):blockLines(6,2)),'%c',80);
catch
    errMessage = ['error reading header data at line ' num2str(lineN) ' relatively to current data-set'];
    return
end


%--------------------------------------------------------------------------
function [UFF,errMessage] = extract164(DATA,blockLines)
% #164 - Extract data-set type 164 data
UFF = [];
errMessage = [];
lineN = 1;
nLines = size(blockLines,1);
try
    % Line 1
    tmpLine = DATA(blockLines(1,1):blockLines(1,2));
    tmpLine = [tmpLine repmat(' ',1,80-length(tmpLine))];
    UFF.unitsCode = sscanf(tmpLine(1:10),'%i');    
    UFF.unitsDescription = sscanf(tmpLine(11:31),'%c');
    UFF.tempMode = sscanf(tmpLine(32:41),'%i');
    lineN = lineN + 1;
    % Line 2
    tmpLine = DATA(blockLines(2,1):blockLines(2,2));
    tmpLine = lower([tmpLine repmat(' ',1,80-length(tmpLine))]);
    tmpLine = strrep(tmpLine,'d-','e-');
    tmpLine = strrep(tmpLine,'d+','e+');
    UFF.facLength = sscanf(tmpLine(1:25),'%f');
    UFF.facForce = sscanf(tmpLine(26:50),'%f');
    UFF.facTemp = sscanf(tmpLine(51:75),'%f');
    lineN = lineN + 1;
    % Line 3
    tmpLine = DATA(blockLines(3,1):blockLines(3,2));
    tmpLine = lower([tmpLine repmat(' ',1,80-length(tmpLine))]);
    tmpLine = strrep(tmpLine,'d-','e-');
    tmpLine = strrep(tmpLine,'d+','e+');
    UFF.facTempOffset = sscanf(tmpLine(1:25),'%f');
catch
    errMessage = ['error reading units data at line' num2str(lineN) ' relatively to current data-set: ' lasterr];
    return
end



%--------------------------------------------------------------------------
function [UFF,errMessage] = extract15(DATA,blockLines)
% #15 - Extract data-set type 15 data

UFF = [];
errMessage = [];
nLines = size(blockLines,1);

try
    values = sscanf(DATA(blockLines(1,1):blockLines(end,2)),'%g');
    nVals = length(values);
    nNodes = round(nVals/7);
    values = reshape(values,7,nNodes).';
    %
    UFF.nodeN = round(values(:,1));
    UFF.defCS = round(values(:,2));
    UFF.dispCS = round(values(:,3));
    UFF.color = round(values(:,4));
    UFF.x = values(:,5);
    UFF.y = values(:,6);
    UFF.z = values(:,7);
catch
    errMessage = ['error reading coordinate data: ' lasterr];
    return
end



%--------------------------------------------------------------------------
function [UFF,errMessage] = extract82(DATA,blockLines)
% #82 - Extract display sequence data-set type 82 data

UFF = [];
errMessage = [];
lineN = 1;
nLines = size(blockLines,1);
try
    % Line 1
    tmpLine = DATA(blockLines(1,1):blockLines(1,2));
    tmpLine = [tmpLine repmat(' ',1,80-length(tmpLine))];
    UFF.traceNum = sscanf(tmpLine(1:10),'%i',1);
    UFF.nNodes = sscanf(tmpLine(11:20),'%i',1);
    UFF.color = sscanf(tmpLine(21:30),'%i',1);
    lineN = lineN + 1;
    % Line 2
    UFF.ID = sscanf(DATA(blockLines(2,1):blockLines(2,2)),'%c');
    lineN = lineN + 1;
    % Line 3
    UFF.lines = sscanf(DATA(blockLines(3,1):blockLines(end,2)),'%g');
catch
    errMessage = ['error reading trace-line data at line' num2str(lineN) ' relatively to current data-set: ' lasterr];
    return
end



%--------------------------------------------------------------------------
function [UFF,errMessage] = extract55(DATA,blockLines)
% #55 - Extract modal data-set type 55 data

UFF = [];
errMessage = [];
lineN = 1;
nLines = size(blockLines,1);
errN = 0;

try
    % Line 1
    UFF.d1 = sscanf(DATA(blockLines(1,1):blockLines(1,2)),'%c',80);
    lineN = lineN + 1;
    % Line 2
    UFF.d2 = sscanf(DATA(blockLines(2,1):blockLines(2,2)),'%c',80);
    lineN = lineN + 1;
    % Line 3
    UFF.date = sscanf(DATA(blockLines(3,1):blockLines(3,2)),'%c',80);
    lineN = lineN + 1;
    % Line 4
    UFF.IDs = sscanf(DATA(blockLines(4,1):blockLines(4,2)),'%c',80);
    % Line 5
    temp = sscanf(DATA(blockLines(5,1):blockLines(5,2)),'%s',1);
    lineN = lineN + 1;
    % Line 6
    tmpLine = DATA(blockLines(6,1):blockLines(6,2));
    tmpLine = [tmpLine repmat(' ',1,80-length(tmpLine))];
    UFF.modelType = sscanf(tmpLine(1:10),'%i',1);
    lineN = lineN + 1;
    if UFF.modelType ~=1,
        errMessage = ['not structural model type (line: ' num2str(lineN) ' relatively to current data-set)'];
        return
    end
    UFF.analysisType = sscanf(tmpLine(11:20),'%i',1);
    UFF.dataCharacter = sscanf(tmpLine(21:30),'%i',1);
    UFF.responseType = sscanf(tmpLine(31:40),'%i',1);
    UFF.dataType = sscanf(tmpLine(41:50),'%i',1);
    num_data_per_pt = sscanf(tmpLine(51:60),'%s',1);

    % Read records 7 and 8 which are analysis-type dependant
    if UFF.analysisType == 2  % Normal Mode
        % Line 7
        tmpLine = DATA(blockLines(7,1):blockLines(7,2));
        tmpLine = [tmpLine repmat(' ',1,80-length(tmpLine))];
        two = sscanf(tmpLine(1:10),'%i',1);
        lineN = lineN + 1;
        if two ~= 2,
            errMessage = ['unexpected value at line ' num2str(lineN) ' relatively to current data-set'];
            return
        end
        four = sscanf(tmpLine(11:20),'%i',1);
        if four ~= 4,
            errMessage = ['unexpected value at line: ' num2str(lineN) ' relatively to current data-set'];
            return
        end
        tmp = sscanf(tmpLine(21:30),'%i',1);
        UFF.modeNum = sscanf(tmpLine(31:40),'%i',1);
        % Line 8
        tmpLine = DATA(blockLines(8,1):blockLines(8,2));
        tmpLine = [tmpLine repmat(' ',1,80-length(tmpLine))];
        lineN = lineN + 1;
        UFF.modeFreq = sscanf(tmpLine(1:13),'%g',1);
        UFF.modeMass = sscanf(tmpLine(14:26),'%g',1);
        UFF.mode_v_damping_ratio = sscanf(tmpLine(27:39),'%g',1);
        UFF.mode_h_damping_ratio = sscanf(tmpLine(40:52),'%g',1);

    elseif UFF.analysisType == 3, % Complex Eigenvalue, First Order (Displacement)
        % Line 7
        tmpLine = DATA(blockLines(7,1):blockLines(7,2));
        tmpLine = [tmpLine repmat(' ',1,80-length(tmpLine))];
        lineN = lineN + 1;
        two = sscanf(tmpLine(1:10),'%i',1);
        if two ~= 2,
            errMessage = ['unexpected value at line ' num2str(lineN) ' relatively to current data-set'];
            return
        end
        six = sscanf(tmpLine(11:20),'%i',1);
        if six ~= 6,
            errMessage = ['unexpected value at line: ' num2str(lineN) ' relatively to current data-set'];
            return
        end
        tmp = sscanf(tmpLine(21:30),'%i',1);
        UFF.modeNum = sscanf(tmpLine(31:40),'%i',1);

        % Line 8
        tmpLine = DATA(blockLines(8,1):blockLines(8,2));
        tmpLine = [tmpLine repmat(' ',1,80-length(tmpLine))];
        lineN = lineN + 1;
        real_part = sscanf(tmpLine(1:13),'%g',1);
        imaginary_part = sscanf(tmpLine(14:26),'%g',1);
        UFF.eigVal = real_part + j * imaginary_part;
        real_part = sscanf(tmpLine(27:39),'%g',1);
        imaginary_part = sscanf(tmpLine(40:52),'%g',1);
        UFF.modalA = real_part + j * imaginary_part;
        real_part = sscanf(tmpLine(53:65),'%g',1);
        imaginary_part = sscanf(tmpLine(66:78),'%g',1);
        UFF.modalB = real_part + j * imaginary_part;

    elseif UFF.analysisType == 5, % Frequency Response
        % Line 7
        tmpLine = DATA(blockLines(7,1):blockLines(7,2));
        tmpLine = [tmpLine repmat(' ',1,80-length(tmpLine))];
        lineN = lineN + 1;
        two = sscanf(tmpLine(1:10),'%i',1);
        if two ~= 2,
            errMessage = ['unexpected value at line ' num2str(lineN) ' relatively to current data-set'];
            return
        end
        one = sscanf(tmpLine(11:20),'%i',1);
        if one ~= 1,            
            errMessage = ['unexpected value at line ' num2str(lineN) ' relatively to current data-set'];
            return
        end
        tmp = sscanf(tmpLine(21:30),'%i',1);
        UFF.freqNum = sscanf(tmpLine(31:40),'%i',1);

        % Line 8
        tmpLine = DATA(blockLines(8,1):blockLines(8,2));
        tmpLine = [tmpLine repmat(' ',1,80-length(tmpLine))];
        lineN = lineN + 1;
        UFF.freq = sscanf(tmpLine(1:13),'%g',1);    % in Hz

    elseif UFF.analysisType == 7  % Complex Eigenvalue, Second Order (Velocity)
        % Line 7
        tmpLine = DATA(blockLines(7,1):blockLines(7,2));
        tmpLine = [tmpLine repmat(' ',1,80-length(tmpLine))];
        lineN = lineN + 1;
        two = sscanf(tmpLine(1:10),'%i',1);
        if two ~= 2,
            errMessage = ['unexpected value at line ' num2str(lineN) ' relatively to current data-set'];
            return
        end
        six = sscanf(tmpLine(11:20),'%i',1);
        if six ~= 6,
            errMessage = ['unexpected value at line ' num2str(lineN) ' relatively to current data-set'];
            return
        end
        tmp = sscanf(tmpLine(21:30),'%i',1);
        UFF.modeNum = sscanf(tmpLine(31:40),'%i',1);

        % Line 8
        tmpLine = DATA(blockLines(8,1):blockLines(8,2));
        tmpLine = [tmpLine repmat(' ',1,80-length(tmpLine))];
        lineN = lineN + 1;
        real_part = sscanf(tmpLine(1:13),'%g',1);
        imaginary_part = sscanf(tmpLine(14:26),'%g',1);
        UFF.eigVal = real_part + j * imaginary_part;
        real_part = sscanf(tmpLine(27:39),'%g',1);
        imaginary_part = sscanf(tmpLine(40:52),'%g',1);
        UFF.modalA = real_part + j * imaginary_part;
        real_part = sscanf(tmpLine(53:65),'%g',1);
        imaginary_part = sscanf(tmpLine(66:78),'%g',1);
        UFF.modalB = real_part + j * imaginary_part;

    else
        errMessage = ['analysis type is not supported at line ' num2str(lineN) ' relatively to current data-set'];
        return
    end

    % Read response data by x,y,...components into r1..r6
    ii = 0;
    nnodes = floor((nLines-9)/2)+1;
    r1 = zeros(nnodes,1);
    r2 = r1;
    r3 = r1;
    r4 = r1;
    r5 = r1;
    r6 = r1;
    nodeNum = r1;
    for lne = 9:2:nLines-1,
        ii = ii + 1;
        %========
        %    line = 9; % line 9 type of line
        %========
        lineRead = lne;
        tmpLine = DATA(blockLines(lineRead,1):blockLines(lineRead,2));
        nodeNum(ii) = sscanf(tmpLine(1:10),'%i',1);
        lineN = lineN + 1;

        %========
        %    line = 10; % line 9 type of line
        %========
        lineRead = lne + 1;
        lineN = lineN + 1;
        tmpLine = DATA(blockLines(lineRead,1):blockLines(lineRead,2));
%         tmpLine = [tmpLine repmat(' ',1,80-length(tmpLine))];
        if UFF.dataType == 2,       % real data
            r1(ii) = sscanf(tmpLine(1:13),'%g',1);
            r2(ii) = sscanf(tmpLine(14:26),'%g',1);
            r3(ii) = sscanf(tmpLine(27:39),'%g',1);
            if num_data_per_pt == 6,
                r4(ii) = sscanf(tmpLine(40:52),'%g',1);
                r5(ii) = sscanf(tmpLine(53:65),'%g',1);
                r6(ii) = sscanf(tmpLine(66:78),'%g',1);
            end
        elseif UFF.dataType == 5,   % complex data
            p1 = sscanf(tmpLine(1:13),'%g',1);
            p2 = sscanf(tmpLine(14:26),'%g',1);
            r1(ii) = p1 + j * p2;
            p3 = sscanf(tmpLine(27:39),'%g',1);
            p4 = sscanf(tmpLine(40:52),'%g',1);
            r2(ii) = p3 + j * p4;
            p5 = sscanf(tmpLine(53:65),'%g',1);
            p6 = sscanf(tmpLine(66:78),'%g',1);
            r3(ii) = p5 + j * p6;
            if num_data_per_pt == 6,
                errMessage = ['not setup to handle six coordinate of complex data at line ' num2str(lineN) ' relatively to current data-set'];
                return
            end
        else
            errMessage = ['what kind of data type was that at line ' num2str(lineN) ' relatively to current data-set?'];
            return
        end
    end
    UFF.r1 = r1;
    UFF.r2 = r2;
    UFF.r3 = r3;
    UFF.r4 = r4;
    UFF.r5 = r5;
    UFF.r6 = r6;
    UFF.nodeNum = nodeNum;
catch
    errMessage = ['error reading modal data: ' lasterr];
    return
end



%--------------------------------------------------------------------------
function [UFF,errMessage] = extract2414(DATA,blockLines)
% #2414 - Extract Analysis Dataset, Structural, Normal Modes
% 2005-07-15 D. Coyne Added function extract2414

UFF = [];
errMessage = [];
lineN = 1;
nLines = size(blockLines,1);
errN = 0;

try
    % Line 1
    tmpLine = DATA(blockLines(1,1):blockLines(1,2));
    tmpLine = [tmpLine repmat(' ',1,80-length(tmpLine))];
    UFF.analysisDatasetLabel = sscanf(tmpLine(1:10),'%i',1);
    lineN = lineN + 1;
    % Line 2
    UFF.analysisDatasetName = sscanf(DATA(blockLines(2,1):blockLines(2,2)),'%c',80);
    lineN = lineN + 1;
    % Line 3
    tmpLine = DATA(blockLines(3,1):blockLines(3,2));
    tmpLine = [tmpLine repmat(' ',1,80-length(tmpLine))];
    UFF.analysisDatasetLocation = sscanf(tmpLine(1:10),'%i',1);
    lineN = lineN + 1;
    % Line 4
    UFF.ID1 = sscanf(DATA(blockLines(4,1):blockLines(4,2)),'%c',80);
    lineN = lineN + 1;
    % Line 5
    UFF.ID2 = sscanf(DATA(blockLines(5,1):blockLines(5,2)),'%c',80);
    lineN = lineN + 1;
    % Line 6
    UFF.ID3 = sscanf(DATA(blockLines(6,1):blockLines(6,2)),'%c',80);
    lineN = lineN + 1;
    % Line 7
    UFF.ID4 = sscanf(DATA(blockLines(7,1):blockLines(7,2)),'%c',80);
    lineN = lineN + 1;
    % Line 8
    UFF.ID5 = sscanf(DATA(blockLines(8,1):blockLines(8,2)),'%c',80);
    lineN = lineN + 1;
    % Line 9
    tmpLine = DATA(blockLines(9,1):blockLines(9,2));
    tmpLine = [tmpLine repmat(' ',1,80-length(tmpLine))];
    UFF.modelType = sscanf(tmpLine(1:10),'%i',1);
    lineN = lineN + 1;
    if UFF.modelType ~=1,
        errMessage = ['not structural model type (line: ' num2str(lineN) ' relatively to current data-set)'];
        return
    end
    UFF.analysisType = sscanf(tmpLine(11:20),'%i',1);
    if UFF.analysisType ~=2,
        errMessage = ['not normal mode analysis type (line: ' num2str(lineN) ' relatively to current data-set)'];
        return
    end
    UFF.dataCharacter = sscanf(tmpLine(21:30),'%i',1);
    UFF.resultType = sscanf(tmpLine(31:40),'%i',1);
    if UFF.resultType ~=8,
        errMessage = ['not displacement result type (line: ' num2str(lineN) ' relatively to current data-set)'];
        return
    end
    UFF.dataType = sscanf(tmpLine(41:50),'%i',1);
    num_data_per_pt = sscanf(tmpLine(51:60),'%s',1);
    UFF.ndval = num_data_per_pt;

    % Read records 10 through 13 which are analysis-type dependent
    if UFF.analysisType == 2  % Normal Mode
        % Line 10
        tmpLine = DATA(blockLines(10,1):blockLines(10,2));
        tmpLine = [tmpLine repmat(' ',1,80-length(tmpLine))];
        UFF.designSetID = sscanf(tmpLine(1:10),'%i',1);
        lineN = lineN + 1;
        UFF.iterationNumber = sscanf(tmpLine(11:20),'%i',1);
        UFF.solutionSetID = sscanf(tmpLine(21:30),'%i',1);
        UFF.boundaryCondition = sscanf(tmpLine(31:40),'%i',1);
        tmp = sscanf(tmpLine(41:50),'%i',1);
        UFF.modeNumber = sscanf(tmpLine(51:60),'%i',1);
        tmp = sscanf(tmpLine(61:70),'%i',1);
        tmp = sscanf(tmpLine(71:80),'%i',1);
        % Line 11
        tmpLine = DATA(blockLines(11,1):blockLines(11,2));
        tmpLine = [tmpLine repmat(' ',1,80-length(tmpLine))];
        lineN = lineN + 1;
        UFF.creationOption = sscanf(tmpLine(1:10),'%i',1);
        UFF.numberRetained = sscanf(tmpLine(11:20),'%i',1);
        % Line 12
        tmpLine = DATA(blockLines(12,1):blockLines(12,2));
        tmpLine = [tmpLine repmat(' ',1,80-length(tmpLine))];
        lineN = lineN + 1;
        tmp = sscanf(tmpLine(1:13),'%g',1);
        UFF.frequency = sscanf(tmpLine(14:26),'%g',1);
        tmp = sscanf(tmpLine(27:39),'%g',1);
        UFF.modalMass = sscanf(tmpLine(40:52),'%g',1);
        UFF.viscousDampingRatio = sscanf(tmpLine(53:65),'%g',1);
        UFF.hystereticDampingRatio = sscanf(tmpLine(66:78),'%g',1);
        % Line 13
        tmpLine = DATA(blockLines(13,1):blockLines(13,2));
        tmpLine = [tmpLine repmat(' ',1,80-length(tmpLine))];
        lineN = lineN + 1;
    else
        errMessage = ['analysis type is not supported at line ' num2str(lineN) ' relatively to current data-set'];
        return
    end
    
    % currently only handles the data class (location) "at nodes"
    if UFF.analysisDatasetLocation ~=1,
        errMessage = ['analysis dataset location (class) is not at nodes (line: ' num2str(lineN) ' relatively to current data-set)'];
        return
    end

    % Lines 14 & 15
    % Read response data by x,y,...components into r1..r6
    ii = 0;
    nnodes = floor((nLines-13)/2);
    r1 = zeros(nnodes,1);
    r2 = r1;
    r3 = r1;
    r4 = r1;
    r5 = r1;
    r6 = r1;
    nodeNum = r1;
    for lne = 14:2:nLines-1,
        ii = ii + 1;
        %========
        %    line = 14; % line 14 type of line
        %========
        lineRead = lne;
        tmpLine = DATA(blockLines(lineRead,1):blockLines(lineRead,2));
        nodeNum(ii) = sscanf(tmpLine(1:10),'%i',1);
        lineN = lineN + 1;

        %========
        %    line = 15; % line 15 type of line
        %========
        lineRead = lne + 1;
        lineN = lineN + 1;
        tmpLine = DATA(blockLines(lineRead,1):blockLines(lineRead,2));
%         tmpLine = [tmpLine repmat(' ',1,80-length(tmpLine))];
        if UFF.dataType == 2,       % real data
            r1(ii) = sscanf(tmpLine(1:13),'%g',1);
            r2(ii) = sscanf(tmpLine(14:26),'%g',1);
            r3(ii) = sscanf(tmpLine(27:39),'%g',1);
%            if num_data_per_pt == 6,
                r4(ii) = sscanf(tmpLine(40:52),'%g',1);
                r5(ii) = sscanf(tmpLine(53:65),'%g',1);
                r6(ii) = sscanf(tmpLine(66:78),'%g',1);
                %            end
        elseif UFF.dataType == 5,   % complex data
            p1 = sscanf(tmpLine(1:13),'%g',1);
            p2 = sscanf(tmpLine(14:26),'%g',1);
            r1(ii) = p1 + j * p2;
            p3 = sscanf(tmpLine(27:39),'%g',1);
            p4 = sscanf(tmpLine(40:52),'%g',1);
            r2(ii) = p3 + j * p4;
            p5 = sscanf(tmpLine(53:65),'%g',1);
            p6 = sscanf(tmpLine(66:78),'%g',1);
            r3(ii) = p5 + j * p6;
            if num_data_per_pt == 6,
                errMessage = ['not setup to handle six coordinate of complex data at line ' num2str(lineN) ' relatively to current data-set'];
                return
            end
        else
            errMessage = ['what kind of data type was that at line ' num2str(lineN) ' relatively to current data-set?'];
            return
        end
    end
    UFF.r1 = r1;
    UFF.r2 = r2;
    UFF.r3 = r3;
    UFF.r4 = r4;
    UFF.r5 = r5;
    UFF.r6 = r6;
    UFF.nodeNum = nodeNum;
catch
    errMessage = ['error reading modal data: ' lasterr];
    return
end
